package at.tuwien.endpoints;

import at.tuwien.test.AbstractUnitTest;
import at.tuwien.api.database.AccessTypeDto;
import at.tuwien.api.database.DatabaseAccessDto;
import at.tuwien.entities.database.Database;
import at.tuwien.entities.database.DatabaseAccess;
import at.tuwien.entities.user.User;
import at.tuwien.exception.*;
import at.tuwien.mapper.AccessMapper;
import at.tuwien.repository.DatabaseRepository;
import at.tuwien.repository.UserRepository;
import at.tuwien.service.AccessService;
import lombok.extern.log4j.Log4j2;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.test.context.support.WithAnonymousUser;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import java.security.Principal;
import java.util.Optional;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

@Log4j2
@SpringBootTest
@ExtendWith(SpringExtension.class)
public class AccessEndpointUnitTest extends AbstractUnitTest {

    @MockBean
    private AccessService accessService;

    @MockBean
    private DatabaseRepository databaseRepository;

    @MockBean
    private UserRepository userRepository;

    @Autowired
    private AccessEndpoint accessEndpoint;

    @Autowired
    private AccessMapper accessMapper;

    @Test
    @WithAnonymousUser
    public void create_anonymous_fails() {

        /* test */
        assertThrows(org.springframework.security.access.AccessDeniedException.class, () -> {
            generic_create(null, USER_2_ID, null, null);
        });
    }

    @Test
    @WithMockUser(username = USER_4_USERNAME)
    public void create_noRoleNoAccess_fails() {

        /* test */
        assertThrows(org.springframework.security.access.AccessDeniedException.class, () -> {
            generic_create(USER_2_PRINCIPAL, USER_4_ID, USER_4_USERNAME, USER_4);
        });
    }

    @Test
    @WithMockUser(username = USER_1_USERNAME, authorities = {"create-database-access"})
    public void create_succeeds() throws ServiceException, ServiceConnectionException, NotAllowedException,
            DatabaseNotFoundException, UserNotFoundException, AccessNotFoundException, SearchServiceException,
            SearchServiceConnectionException {

        /* mock */
        doNothing()
                .when(accessService)
                .create(eq(DATABASE_1), eq(USER_2), any(AccessTypeDto.class));

        /* test */
        generic_create(USER_2_PRINCIPAL, USER_2_ID, USER_2_USERNAME, USER_2);
    }

    @Test
    @WithMockUser(username = USER_1_USERNAME, authorities = {"check-database-access"})
    public void find_hasRoleNoAccess_fails() {

        /* test */
        assertThrows(AccessNotFoundException.class, () -> {
            generic_find(DATABASE_1_ID, DATABASE_1, null, USER_2_PRINCIPAL, USER_2_ID, USER_2);
        });
    }

    @Test
    @WithMockUser(username = USER_1_USERNAME, authorities = {"check-database-access"})
    public void find_hasRoleHasAccess_succeeds() throws UserNotFoundException, DatabaseNotFoundException,
            AccessNotFoundException, NotAllowedException {

        /* test */
        generic_find(DATABASE_1_ID, DATABASE_1, DATABASE_1_USER_1_READ_ACCESS, USER_1_PRINCIPAL, USER_1_ID, USER_1);
    }

    @Test
    @WithMockUser(username = USER_1_USERNAME, authorities = {"check-database-access"})
    public void find_hasRoleHasAccessForeign_fails() {

        /* test */
        assertThrows(NotAllowedException.class, () -> {
            generic_find(DATABASE_1_ID, DATABASE_1, DATABASE_1_USER_1_READ_ACCESS, USER_1_PRINCIPAL, USER_2_ID, USER_2);
        });
    }

    @Test
    @WithMockUser(username = USER_1_USERNAME, authorities = {"admin"})
    public void find_hasRoleHasAccessForeign_succeeds() throws UserNotFoundException, NotAllowedException,
            DatabaseNotFoundException, AccessNotFoundException {

        /* test */
        generic_find(DATABASE_1_ID, DATABASE_1, DATABASE_1_USER_1_READ_ACCESS, USER_LOCAL_ADMIN_PRINCIPAL, USER_1_ID, USER_1);
    }

    @Test
    @WithAnonymousUser
    public void update_anonymous_fails() {

        /* test */
        assertThrows(org.springframework.security.access.AccessDeniedException.class, () -> {
            generic_update(null, USER_4_USERNAME, USER_4, null, null);
        });
    }

    @Test
    @WithMockUser(username = USER_1_USERNAME, authorities = {"update-database-access"})
    public void update_hasRoleNoAccess_fails() {

        /* test */
        assertThrows(NotAllowedException.class, () -> {
            generic_update(null, USER_4_USERNAME, USER_4, USER_1_PRINCIPAL, USER_1);
        });
    }

    @Test
    @WithMockUser(username = USER_4_USERNAME)
    public void update_noRoleNoAccess_fails() {

        /* test */
        assertThrows(org.springframework.security.access.AccessDeniedException.class, () -> {
            generic_update(null, USER_4_USERNAME, USER_4, USER_4_PRINCIPAL, USER_4);
        });
    }

    @Test
    @WithMockUser(username = USER_1_USERNAME, authorities = {"update-database-access"})
    public void update_succeeds() throws NotAllowedException, ServiceException, ServiceConnectionException,
            AccessNotFoundException, DatabaseNotFoundException, UserNotFoundException, SearchServiceException,
            SearchServiceConnectionException {

        /* mock */
        doNothing()
                .when(accessService)
                .update(eq(DATABASE_1), eq(USER_2), any(AccessTypeDto.class));

        /* test */
        generic_update(DATABASE_1_USER_2_WRITE_OWN_ACCESS, USER_2_USERNAME, USER_2, USER_2_PRINCIPAL, USER_2);
    }

    @Test
    @WithAnonymousUser
    public void revoke_anonymous_fails() {

        /* test */
        assertThrows(org.springframework.security.access.AccessDeniedException.class, () -> {
            generic_revoke(USER_1_PRINCIPAL, USER_1);
        });
    }

    @Test
    @WithMockUser(username = USER_4_USERNAME)
    public void revoke_noRoleNoAccess_fails() {

        /* test */
        assertThrows(org.springframework.security.access.AccessDeniedException.class, () -> {
            generic_revoke(USER_4_PRINCIPAL, USER_4);
        });
    }

    @Test
    @WithMockUser(username = USER_1_USERNAME, authorities = {"delete-database-access"})
    public void revoke_succeeds() throws NotAllowedException, ServiceException, ServiceConnectionException,
            UserNotFoundException, DatabaseNotFoundException, AccessNotFoundException, SearchServiceException,
            SearchServiceConnectionException {

        /* mock */
        doNothing()
                .when(accessService)
                .delete(DATABASE_1, USER_2);

        /* test */
        generic_revoke(USER_1_PRINCIPAL, USER_1);
    }

    /* ################################################################################################### */
    /* ## GENERIC TEST CASES                                                                            ## */
    /* ################################################################################################### */

    protected void generic_create(Principal principal, UUID userId, String username, User user)
            throws NotAllowedException, ServiceException, ServiceConnectionException, UserNotFoundException,
            DatabaseNotFoundException, AccessNotFoundException, SearchServiceException,
            SearchServiceConnectionException {

        /* mock */
        when(databaseRepository.findById(DATABASE_1_ID))
                .thenReturn(Optional.of(DATABASE_1));
        doThrow(AccessNotFoundException.class)
                .when(accessService)
                .find(DATABASE_1, user);
        if (user != null) {
            when(userRepository.findByUsername(username))
                    .thenReturn(Optional.of(user));
        } else {
            when(userRepository.findByUsername(anyString()))
                    .thenReturn(Optional.empty());
        }

        /* test */
        final ResponseEntity<?> response = accessEndpoint.create(DATABASE_1_ID, userId, UPDATE_DATABASE_ACCESS_READ_DTO, principal);
        assertEquals(HttpStatus.ACCEPTED, response.getStatusCode());
        assertNull(response.getBody());
    }

    protected void generic_find(Long databaseId, Database database, DatabaseAccess access, Principal principal,
                                UUID userId, User user) throws UserNotFoundException, DatabaseNotFoundException,
            AccessNotFoundException, NotAllowedException {

        /* mock */
        when(databaseRepository.findById(databaseId))
                .thenReturn(Optional.of(database));
        when(userRepository.findById(userId))
                .thenReturn(Optional.of(user));
        if (access != null) {
            log.trace("mock access {} for user with id {} for database with id {}", access.getType(), userId, databaseId);
            when(accessService.find(database, user))
                    .thenReturn(access);
        } else {
            log.trace("mock no access for user with id {} for database with id {}", userId, databaseId);
            doThrow(AccessNotFoundException.class)
                    .when(accessService)
                    .find(database, user);
        }
        if (principal != null) {
            when(userRepository.findByUsername(principal.getName()))
                    .thenReturn(Optional.of(user));
        }

        /* test */
        final ResponseEntity<DatabaseAccessDto> response = accessEndpoint.find(databaseId, userId, principal);
        assertEquals(HttpStatus.OK, response.getStatusCode());
        assertNotNull(response.getBody());
        final DatabaseAccessDto dto = response.getBody();
        assertEquals(userId, dto.getHuserid());
        assertEquals(databaseId, dto.getHdbid());
        assertEquals(accessMapper.accessType(access.getType()), dto.getType());
    }

    protected void generic_update(DatabaseAccess access, String otherUsername, User otherUser, Principal principal,
                                  User user) throws NotAllowedException, ServiceException, ServiceConnectionException,
            AccessNotFoundException, UserNotFoundException, DatabaseNotFoundException, SearchServiceException,
            SearchServiceConnectionException {

        /* mock */
        when(databaseRepository.findById(DATABASE_1_ID))
                .thenReturn(Optional.of(DATABASE_1));
        if (access != null) {
            log.trace("mock access {} for user with id {} for database with id {}", access.getType(), USER_1_ID, DATABASE_1_ID);
            when(accessService.find(DATABASE_1, USER_1))
                    .thenReturn(access);
        } else {
            log.trace("mock no access for user with id {} for database with id {}", USER_1_ID, DATABASE_1_ID);
            doThrow(AccessNotFoundException.class)
                    .when(accessService)
                    .find(DATABASE_1, USER_1);
        }
        if (otherUsername != null) {
            when(userRepository.findByUsername(otherUsername))
                    .thenReturn(Optional.of(otherUser));
        } else {
            when(userRepository.findByUsername(anyString()))
                    .thenReturn(Optional.empty());
        }
        if (principal != null) {
            when(userRepository.findByUsername(principal.getName()))
                    .thenReturn(Optional.of(user));
        } else {
            when(userRepository.findByUsername(anyString()))
                    .thenReturn(Optional.empty());
        }

        /* test */
        final ResponseEntity<?> response = accessEndpoint.update(DATABASE_1_ID, USER_1_ID, UPDATE_DATABASE_ACCESS_READ_DTO, principal);
        assertEquals(HttpStatus.ACCEPTED, response.getStatusCode());
        assertNull(response.getBody());
    }

    protected void generic_revoke(Principal principal, User user) throws ServiceConnectionException,
            NotAllowedException, ServiceException, UserNotFoundException, DatabaseNotFoundException,
            AccessNotFoundException, SearchServiceException, SearchServiceConnectionException {

        /* mock */
        when(accessService.find(any(Database.class), eq(user)))
                .thenReturn(DATABASE_1_USER_1_READ_ACCESS);
        when(databaseRepository.findById(DATABASE_1_ID))
                .thenReturn(Optional.of(DATABASE_1));
        if (principal != null) {
            when(userRepository.findByUsername(principal.getName()))
                    .thenReturn(Optional.of(user));
        }

        /* test */
        final ResponseEntity<?> response = accessEndpoint.revoke(DATABASE_1_ID, USER_1_ID, principal);
        assertEquals(HttpStatus.ACCEPTED, response.getStatusCode());
        assertNull(response.getBody());
    }

}
