package at.tuwien.service.impl;

import at.tuwien.api.auth.SignupRequestDto;
import at.tuwien.api.user.*;
import at.tuwien.entities.user.User;
import at.tuwien.exception.*;
import at.tuwien.repository.UserRepository;
import at.tuwien.service.UserService;
import lombok.extern.log4j.Log4j2;
import org.apache.commons.codec.digest.DigestUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

@Log4j2
@Service
public class UserServiceImpl implements UserService {

    private final UserRepository userRepository;

    @Autowired
    public UserServiceImpl(UserRepository userRepository) {
        this.userRepository = userRepository;
    }

    @Override
    public List<User> findAll() {
        return userRepository.findAll();
    }

    @Override
    public User findByUsername(String username) throws UserNotFoundException {
        final Optional<User> optional = userRepository.findByUsername(username);
        if (optional.isEmpty()) {
            log.error("Failed to find user with username {}", username);
            throw new UserNotFoundException("Failed to find user with username " + username);
        }
        return optional.get();
    }

    @Override
    public User findById(UUID id) throws UserNotFoundException {
        final Optional<User> optional = userRepository.findById(id);
        if (optional.isEmpty()) {
            log.error("Failed to find user with id {}", id);
            throw new UserNotFoundException("Failed to find user with id " + id);
        }
        return optional.get();
    }

    @Override
    public User create(SignupRequestDto data, UUID id) {
        /* create at authentication service */
        final User entity = User.builder()
                .id(id)
                .username(data.getUsername())
                .email(data.getEmail())
                .theme("light")
                .mariadbPassword(getMariaDbPassword(data.getPassword()))
                .language("en")
                .build();
        /* create at metadata database */
        final User user = userRepository.save(entity);
        log.info("Created user with id {}", user.getId());
        return user;
    }

    @Override
    public User modify(User user, UserUpdateDto data) {
        user.setFirstname(data.getFirstname());
        user.setLastname(data.getLastname());
        user.setAffiliation(data.getAffiliation());
        user.setOrcid(data.getOrcid());
        user.setTheme(data.getTheme());
        user.setLanguage(data.getLanguage());
        /* create at metadata database */
        user = userRepository.save(user);
        log.info("Modified user with id {}", user.getId());
        return user;
    }

    @Override
    public void updatePassword(User user, UserPasswordDto data) {
        user.setMariadbPassword(getMariaDbPassword(data.getPassword()));
        /* update at metadata database */
        userRepository.save(user);
        log.info("Updated password of user with id {}", user.getId());
    }

    @Override
    public void validateUsernameNotExists(String username) throws UserExistsException {
        if (userRepository.existsByUsername(username)) {
            throw new UserExistsException("User with username " + username + " already exists");
        }
    }

    @Override
    public void validateEmailNotExists(String email) throws EmailExistsException {
        if (userRepository.existsByEmail(email)) {
            throw new EmailExistsException("User with email " + email + " already exists");
        }
    }

    public String getMariaDbPassword(String password) {
        final byte[] utf8 = password.getBytes(StandardCharsets.UTF_8);
        return "*" + DigestUtils.sha1Hex(DigestUtils.sha1(utf8)).toUpperCase();
    }
}
